/**
 * This file is yet under development. Please use at your own risk.
 *
 * Zerjillo
 */
package com.zerjio.JIRC;

import java.io.*;
import java.util.*;

/**
 * This class allows a higher level of interaction within the IRC client and
 * the IRC server. This class is the cornerstone of the package. Every application
 * which wants to communicate with the IRC server should use this class. It allows
 * sending the basic IRC commands, and even a RAW command (a command that it is
 * not directly implemented).
 *
 * It also is responsible of reading all incoming IRC messages from the server and
 * to deliver it to the appropiate message listeners.
 *
 * @author  Zerjillo (http://www.zerjio.com)
 * @version 0.9b
 */
public class IRCClient implements Runnable {
  /** It will check every 200 ms for incoming messages */
  private static final int READING_INTERVAL = 200;
  
  /** Server name */
  private String server;
  /** Port of the server where the client will connect */
  private int port;
  
  /** The connection to the server */
  private IRCConnection connection;
  /** Marks when the client should stop reading messages from server */
  private boolean stopReading;
  /** The thread that reads messages from the server */
  private Thread readThread;
  
  /** List of all <code>MsgListener</code>s */
  private Vector msgListeners;
  /** List of all <code>PINGListener</code>s */
  private Vector pingListeners;
  /** List of all <code>NICKListener</code>s */
  private Vector nickListeners;
  /** List of all <code>ReplyMsgListener</code>s */
  private Vector replyMsgListeners;
  /** List of all <code>QUITListener</code>s */
  private Vector quitListeners;
  /** List of all <code>JOINListener</code>s */
  private Vector joinListeners;
  /** List of all <code>PARTListener</code>s */
  private Vector partListeners;
  /** List of all <code>TOPICListener</code>s */
  private Vector topicListeners;
  /** List of all <code>PRIVMSGListener</code>s */
  private Vector privmsgListeners;  
  
  /**
   * Creates a new IRC client without specifying a server nor port to connect
   */
  public IRCClient() {
    server = "";
    port = -1;
    msgListeners = new Vector();
    pingListeners = new Vector();
    nickListeners = new Vector();
    replyMsgListeners = new Vector();
    quitListeners = new Vector();
    joinListeners = new Vector();
    partListeners = new Vector();
    topicListeners = new Vector();
    privmsgListeners = new Vector();
  }
  
  /**
   * Creates a new IRC client with a determined server name and port.
   *
   * @param   server    The server name (or IP adress)
   * @param   port      The server port where the client will connect.
   */
  public IRCClient(String server, int port) {
    this.server = server;
    this.port = port;
    msgListeners = new Vector();
    pingListeners = new Vector();
    nickListeners = new Vector();
    replyMsgListeners = new Vector();
    quitListeners = new Vector();
    joinListeners = new Vector();
    partListeners = new Vector();
    topicListeners = new Vector();
    privmsgListeners = new Vector();
  }
  
  /**
   * Returns the server name
   *
   * @return  The server name
   */
  public String getServer() {
    return server;
  }
  
  /**
   * Returns the server port where the client will connect.
   *
   * @return The server port where the client will connect.
   */
  public int getPort() {
    return port;
  }
  
  /**
   * Sets the server name
   *
   * @param   server    The server name
   */
  public void setServer(String server) {
    this.server = server;
  }
  
  /**
   * Sets the server port where will connect the client.
   *
   * @param   port    The server port
   */
  public void setPort(int port) {
    this.port = port;
  }
  
  /**
   * Sets the server name and port where will connect the client.
   *
   * @param   server    The server name
   * @param   port      The server port
   */
  public void setServerAndPort(String server, int port) {
    this.server = server;
    this.port = port;
  }
  
  /**
   * Makes a connection to the IRC server. It also starts listening to messages
   * from the server. If there was a previous connection it will be closed first.
   *
   * @return    <code>true</code> if the connection is successful.
   *            <code>false</code> if the conection has failed.
   */
  public boolean connect() {
    disconnect();
    
    //    System.err.print("Connecting to " + server + ":" + port + "... ");
    
    try {
      connection = new IRCConnection(server, port);
    } catch(IOException e) {
      System.err.println("Error when connecting.");
      e.printStackTrace(System.err);
      return false;
    }
    
    readThread = new Thread(this);
    readThread.start();
    
    //    System.err.println("Connection was successful.");
    return true;
  }
  
  /**
   * Returns if the connection is closed.
   *
   * @return    <code>true</code> if the connection is closed.
   *            <code>false</code> if the connection is active.
   */
  public boolean isClosed() {
    if (connection != null) {
      return connection.isClosed();
    } else {
      return true;
    }
  }
  
  /**
   * Breaks the connection with the IRC server. It also stops reading messages from it.
   */
  public void disconnect() {
    if (connection != null) {
      if (!connection.isClosed()) {
        //        System.err.print("Disconnecting from IRC Server... ");
        stopReading = true;
        
        try {
          Thread.sleep(READING_INTERVAL);
        } catch (InterruptedException e) {}
        
        try {
          connection.disconnect();
          //          System.err.println("Disconnection successful.");
        } catch(IOException e) {
          System.err.println("Error when disconnecting.");
          e.printStackTrace(System.err);
        }
      }
      
      connection = null;
    }
  }
  
  /**
   * Adds a <code>PINGListener</code>. It will be placed in the last position between active
   * <code>PINGListener</code>s. When a <code>PINGMessage</code> is received it will be
   * delivered to all <code>PINGListener</code>s (see <code>processRawMessage(...)</code> for more details).
   *
   * @param   listener    The <code>PINGListener</code> to be added.
   */
  public void addPINGListener(PINGListener listener) {
    pingListeners.add(listener);
  }
  
  /**
   * Removes a <code>PINGListener</code> that has been previously added.
   *
   * @param   listener    The <code>PINGListener</code> that should be removed.
   * @see     #addPINGListener(PINGListener listener) addPINGListener
   */
  public void removePINGListener(PINGListener listener) {
    pingListeners.remove(listener);
  }
  
  /**
   * Removes all <code>PINGListener</code>s that have been previously added.
   *
   * @see     #addPINGListener(PINGListener listener) addPINGListener
   */
  public void removeAllPINGListeners() {
    pingListeners.removeAllElements();
  }

  /**
   * Adds a <code>PRIVMSGListener</code>. It will be placed in the last position between active
   * <code>PRIVMSGListener</code>s. When a <code>PRIVMSGMessage</code> is received it will be
   * delivered to all <code>PRIVMSGListener</code>s (see <code>processRawMessage(...)</code> for more details).
   *
   * @param   listener    The <code>PRIVMSGListener</code> to be added.
   */
  public void addPRIVMSGListener(PRIVMSGListener listener) {
    privmsgListeners.add(listener);
  }
  
  /**
   * Removes a <code>PRIVMSGListener</code> that has been previously added.
   *
   * @param   listener    The <code>PRIVMSGListener</code> that should be removed.
   * @see     #addPRIVMSGListener(PRIVMSGListener listener) addPRIVMSGListener
   */
  public void removePRIVMSGListener(PRIVMSGListener listener) {
    privmsgListeners.remove(listener);
  }
  
  /**
   * Removes all <code>PRIVMSGListener</code>s that have been previously added.
   *
   * @see     #addPRIVMSGListener(PRIVMSGListener listener) addPRIVMSGListener
   */
  public void removeAllPRIVMSGListeners() {
    privmsgListeners.removeAllElements();
  }
  
  /**
   * Adds a <code>TOPICListener</code>. It will be placed in the last position between active
   * <code>TOPICListener</code>s. When a <code>TOPICMessage</code> is received it will be
   * delivered to all <code>TOPICListener</code>s (see <code>processRawMessage(...)</code> for more details).
   *
   * @param   listener    The <code>TOPICListener</code> to be added.
   */
  public void addTOPICListener(TOPICListener listener) {
    topicListeners.add(listener);
  }
  
  /**
   * Removes a <code>TOPICListener</code> that has been previously added.
   *
   * @param   listener    The <code>TOPICListener</code> that should be removed.
   * @see     #addTOPICListener(TOPICListener listener) addTOPICListener
   */
  public void removeTOPICListener(TOPICListener listener) {
    topicListeners.remove(listener);
  }
  
  /**
   * Removes all <code>TOPICListener</code>s that have been previously added.
   *
   * @see     #addTOPICListener(TOPICListener listener) addTOPICListener
   */
  public void removeAllTOPICListeners() {
    topicListeners.removeAllElements();
  }
  
  /**
   * Adds a <code>PARTListener</code>. It will be placed in the last position between active
   * <code>PARTListener</code>s. When a <code>PARTMessage</code> is received it will be
   * delivered to all <code>PARTListener</code>s (see <code>processRawMessage(...)</code> for more details).
   *
   * @param   listener    The <code>PARTListener</code> to be added.
   */
  public void addPARTListener(PARTListener listener) {
    partListeners.add(listener);
  }
  
  /**
   * Removes a <code>PARTListener</code> that has been previously added.
   *
   * @param   listener    The <code>PARTListener</code> that should be removed.
   * @see     #addPARTListener(PARTListener listener) addPARTListener
   */
  public void removePARTListener(PARTListener listener) {
    partListeners.remove(listener);
  }
  
  /**
   * Removes all <code>PARTListener</code>s that have been previously added.
   *
   * @see     #addPARTListener(PARTListener listener) addPARTListener
   */
  public void removeAllPARTListeners() {
    partListeners.removeAllElements();
  }
  
  /**
   * Adds a <code>JOINListener</code>. It will be placed in the last position between active
   * <code>JOINListener</code>s. When a <code>JOINMessage</code> is received it will be
   * delivered to all <code>JOINListener</code>s (see <code>processRawMessage(...)</code> for more details).
   *
   * @param   listener    The <code>JOINListener</code> to be added.
   */
  public void addJOINListener(JOINListener listener) {
    joinListeners.add(listener);
  }
  
  /**
   * Removes a <code>JOINListener</code> that has been previously added.
   *
   * @param   listener    The <code>JOINListener</code> that should be removed.
   * @see     #addJOINListener(JOINListener listener) addJOINListener
   */
  public void removeJOINListener(JOINListener listener) {
    joinListeners.remove(listener);
  }
  
  /**
   * Removes all <code>JOINListener</code>s that have been previously added.
   *
   * @see     #addJOINListener(JOINListener listener) addJOINListener
   */
  public void removeAllJOINListeners() {
    joinListeners.removeAllElements();
  }
  
  /**
   * Adds a <code>QUITListener</code>. It will be placed in the last position between active
   * <code>QUITListener</code>s. When a <code>QUITMessage</code> is received it will be
   * delivered to all <code>QUITListener</code>s (see <code>processRawMessage(...)</code> for more details).
   *
   * @param   listener    The <code>QUITListener</code> to be added.
   */
  public void addQUITListener(QUITListener listener) {
    quitListeners.add(listener);
  }
  
  /**
   * Removes a <code>QUITListener</code> that has been previously added.
   *
   * @param   listener    The <code>QUITListener</code> that should be removed.
   * @see     #addQUITListener(QUITListener listener) addQUITListener
   */
  public void removeQUITListener(QUITListener listener) {
    quitListeners.remove(listener);
  }
  
  /**
   * Removes all <code>QUITListener</code>s that have been previously added.
   *
   * @see     #addQUITListener(QUITListener listener) addQUITListener
   */
  public void removeAllQUITListeners() {
    quitListeners.removeAllElements();
  }
  
  /**
   * Adds a <code>ReplyMsgListener</code>. It will be placed in the last position between active
   * <code>ReplyMsgListener</code>s. When a <code>ReplyMessage</code> is received it will be
   * delivered to all <code>ReplyMsgListener</code>s (see <code>processRawMessage(...)</code> for more details).
   *
   * @param   listener    The <code>ReplyMsgListener</code> to be added.
   */
  public void addReplyMsgListener(ReplyMsgListener listener) {
    replyMsgListeners.add(listener);
  }
  
  /**
   * Removes a <code>ReplyMsgListener</code> that has been previously added.
   *
   * @param   listener    The <code>ReplyMsgListener</code> that should be removed.
   * @see     #addReplyMsgListener(ReplyMsgListener listener) addReplyMsgListener
   */
  public void removeReplyMsgListener(ReplyMsgListener listener) {
    replyMsgListeners.remove(listener);
  }
  
  /**
   * Removes all <code>ReplyMsgListener</code>s that have been previously added.
   *
   * @see     #addReplyMsgListener(ReplyMsgListener listener) addReplyMsgListener
   */
  public void removeAllReplyMsgListeners() {
    replyMsgListeners.removeAllElements();
  }
  
  /**
   * Adds a <code>NICKListener</code>. It will be placed in the last position between active
   * <code>NICKListener</code>s. When a <code>NICKMessage</code> is received it will be
   * delivered to all <code>NICKListener</code>s (see <code>processRawMessage(...)</code> for more details).
   *
   * @param   listener    The <code>NICKListener</code> to be added.

   */
  public void addNICKListener(NICKListener listener) {
    nickListeners.add(listener);
  }
  
  /**
   * Removes a <code>NICKListener</code> that has been previously added.
   *
   * @param   listener    The <code>NICKListener</code> that should be removed.
   * @see     #addNICKListener(NICKListener listener) addNICKListener
   */
  public void removeNICKListener(NICKListener listener) {
    nickListeners.remove(listener);
  }
  
  /**
   * Removes all <code>NICKListener</code>s that have been previously added.
   *
   * @see     #addNICKListener(NICKListener listener) addNICKListener
   */
  public void removeAllNICKListeners() {
    nickListeners.removeAllElements();
  }
  
  /**
   * Adds a <code>MsgListener</code>. It will be placed in the last position between active
   * <code>MsgListener</code>s. When a <code>IRCMessage</code> is received it will be
   * delivered to all <code>MsgListener</code>s (see <code>processRawMessage(...)</code> for more details).
   *
   * @param   listener    The <code>MsgListener</code> to be added.
   */
  public void addMsgListener(MsgListener listener) {
    msgListeners.add(listener);
  }
  
  /**
   * Removes a <code>MsgListener</code> that has been previously added.
   *
   * @param   listener    The <code>MsgListener</code> that should be removed.
   * @see     #addMsgListener(MsgListener listener) addMsgListener
   */
  public void removeMsgListener(MsgListener listener) {
    msgListeners.remove(listener);
  }
  
  /**
   * Removes all <code>MsgListener</code>s that have been previously added.
   *
   * @see     #addMsgListener(MsgListener listener) addMsgListener
   */
  public void removeAllMsgListeners() {
    msgListeners.removeAllElements();
  }
  
  /**
   * Handles a raw  Message (sends it to the apropiate listeners).
   * Do not use it directly.
   *
   * @param   rawMsg    A raw message
   */
  private void processRawMessage(String rawMsg) {
    String prefix = "";
    String command = "";
    String params = "";
    
    if (rawMsg.startsWith(":")) {
      int spacePos = rawMsg.indexOf(" ");
      prefix = rawMsg.substring(0, spacePos);
      rawMsg = rawMsg.substring(spacePos + 1);
    }
    
    int spacePos = rawMsg.indexOf(" ");
    command = rawMsg.substring(0, spacePos);
    params = rawMsg.substring(spacePos + 1);
    
    command = command.toUpperCase();
    
    IRCMessage msg = null;
    
    if (command.compareTo("PING") == 0) {
      msg = new PINGMessage(prefix, params);
      
      for (int i = 0 ; i < pingListeners.size() ; i++) {
        PINGListener listener = (PINGListener)pingListeners.elementAt(i);
        
        listener.PINGReceived((PINGMessage)msg);
        
        if (msg.getMustBeDiscarded()) {
          return;
        }
      }
    } else if (command.compareTo("JOIN") == 0) {
      String[] channels = Utils.splitString(',', params);
      
      if (channels.length == 1) {  // If there is only one channel to join
        msg = new JOINMessage(prefix, params);
        
        for (int i = 0 ; i < joinListeners.size() ; i++) {
          JOINListener listener = (JOINListener)joinListeners.elementAt(i);
          
          if (listener.listensToJOIN(((JOINMessage)msg).getChannelName())) {
            listener.JOINReceived((JOINMessage)msg);
            
            if (msg.getMustBeDiscarded()) {
              return;
            }
          }
        }
      } else {  // If there is more than one channel a recursive call is made to join every channel
        for (int i = 0 ; i < channels.length ; i++) {
          processRawMessage(prefix + " JOIN " + channels[i]);
        }
      }
    } else if (command.compareTo("PART") == 0) {
      String[] channels = Utils.splitString(',', params);
      
      if (channels.length == 1) {  // If there is only one channel to part
        msg = new PARTMessage(prefix, params);
        
        for (int i = 0 ; i < partListeners.size() ; i++) {
          PARTListener listener = (PARTListener)partListeners.elementAt(i);
          
          if (listener.listensToPART(((PARTMessage)msg).getChannelName())) {
            listener.PARTReceived((PARTMessage)msg);
            
            if (msg.getMustBeDiscarded()) {
              return;
            }
          }
        }
      } else {  // If there is more than one channel a recursive call is made to part every channel
        for (int i = 0 ; i < channels.length ; i++) {
          processRawMessage(prefix + " PART " + channels[i]);
        }
      }
    } else if (command.compareTo("PRIVMSG") == 0) {
      String[] parts = Utils.splitString(' ', params);
      String[] recipients = Utils.splitString(',', parts[0]);
      String realMsg = parts[1];
      
      if (recipients.length == 1) {  // If there is only one recipient
        msg = new PRIVMSGMessage(prefix, params);
        
        for (int i = 0 ; i < privmsgListeners.size() ; i++) {
          PRIVMSGListener listener = (PRIVMSGListener)privmsgListeners.elementAt(i);
          
          if (listener.listensToPRIVMSG(((PRIVMSGMessage)msg).getRecipientName())) {
            listener.PRIVMSGReceived((PRIVMSGMessage)msg);
            
            if (msg.getMustBeDiscarded()) {
              return;
            }
          }
        }
      } else {  // If there is more than one recipient a recursive call is made to join every channel
        for (int i = 0 ; i < recipients.length ; i++) {
          processRawMessage(prefix + " PRIVMSG " + recipients[i] + " " + realMsg);
        }
      }
    } else if (command.compareTo("NICK") == 0) {
      msg = new NICKMessage(prefix, params);
      
      for (int i = 0 ; i < nickListeners.size() ; i++) {
        NICKListener listener = (NICKListener)nickListeners.elementAt(i);
        
        listener.NICKReceived((NICKMessage)msg);
        
        if (msg.getMustBeDiscarded()) {
          return;
        }
      }
    } else if (command.compareTo("TOPIC") == 0) {
      msg = new TOPICMessage(prefix, params);
      
      for (int i = 0 ; i < topicListeners.size() ; i++) {
        TOPICListener listener = (TOPICListener)topicListeners.elementAt(i);
        
        if (listener.listensToTOPIC(((TOPICMessage)msg).getChannelName())) {
          listener.TOPICReceived((TOPICMessage)msg);
          
          if (msg.getMustBeDiscarded()) {
            return;
          }
        }
      }
    } else if (command.compareTo("QUIT") == 0) {
      msg = new QUITMessage(prefix, params);
      
      for (int i = 0 ; i < quitListeners.size() ; i++) {
        QUITListener listener = (QUITListener)nickListeners.elementAt(i);
        
        listener.QUITReceived((QUITMessage)msg);
        
        if (msg.getMustBeDiscarded()) {
          return;
        }
      }
    } else if (Utils.isReplyMessage(command)) {
      msg = new ReplyMessage(prefix, command, params);
      
      for (int i = 0 ; i < replyMsgListeners.size() ; i++) {
        ReplyMsgListener listener = (ReplyMsgListener)replyMsgListeners.elementAt(i);
        
        if (listener.listensTo(command)) {
          listener.replyMsgReceived((ReplyMessage)msg);
          
          if (msg.getMustBeDiscarded()) {
            return;
          }
        }
      }
    } else {
      msg = new IRCMessage(prefix, command, params);
    }
    
    for (int i = 0 ; i < msgListeners.size() ; i++) {
      MsgListener listener = (MsgListener)msgListeners.elementAt(i);
      
      listener.msgReceived(msg);
      
      if (msg.getMustBeDiscarded()) {
        return;
      }
    }
  }
  
  /**
   * Gets all information send by the IRC server continuously. It reads from the
   * connection every <code>READING_INTERVAL</code>.
   */
  public void run() {
    String rawMsgs = "";
    stopReading = false;
    while(!stopReading) {
      byte[] readed = null;
      
      if (connection != null) {
        if (!connection.isClosed()) {
          readed = connection.read();
        } else {
          System.err.println("Error reading: Connection is closed");
          stopReading = true;
          connection = null;
        }
      }
      
      if (readed != null) {
        rawMsgs += new String(readed);
        
        int newLinePosition = rawMsgs.indexOf("\r\n");
        
        while(newLinePosition != -1) {
          String rawMsg = rawMsgs.substring(0, newLinePosition);
          
          processRawMessage(rawMsg);
          
          rawMsgs = rawMsgs.substring(newLinePosition + 2);
          newLinePosition = rawMsgs.indexOf("\r\n");
        }
      }
      
      try {
        Thread.sleep(READING_INTERVAL);
      } catch(InterruptedException e) {}
    }
  }
  
  /**
   * Sends a PASS message to the IRC server. The password will be the connection
   * password. If no explicit password is provided it will generate a random one.
   *
   * @param   passwd    The connection password. If <code>null</code> or <code>""</code>
   *                    it will generate a random password.
   */
  public void sendPASSMessage(String passwd) {
    if (passwd == null) {
      sendPASSMessage();
    }
    
    if (passwd.compareTo("") == 0) {
      sendPASSMessage();
    }
    
    sendRawMessage("PASS " + passwd);
  }
  
  /**
   * Sends a PASS message to the IRC server with a random connection password.
   */
  public void sendPASSMessage() {
    sendRawMessage("PASS " + Integer.toString((int)(Math.random() * 1000000), 30));
  }
  
  /**
   * Sends a NICK message to the IRC server.
   *
   * @param   nick    The nick.
   */
  public void sendNICKMessage(String nick) {
    sendRawMessage("NICK " + nick);
  }
  
  /**
   * Sends a USER message to the IRC server.
   *
   * @param   userName    The user name.
   * @param   hostName    The host name.
   * @param   serverName  The server name.
   * @param   realName    The real name.
   */
  public void sendUSERMessage(String userName, String hostName, String serverName, String realName) {
    sendRawMessage("USER " + userName + " " + hostName + " " + serverName + " :" + realName);
  }
  
  /**
   * Sends a QUIT message with an optional text to the IRC server.
   *
   * @param   quitMsg   The quit message (the optional text).
   */
  public void sendQUITMessage(String quitMsg) {
    sendRawMessage("QUIT :" + quitMsg);
  }
  
  /**
   * Sends a QUIT message to the IRC server.
   */
  public void sendQUITMessage() {
    sendRawMessage("QUIT");
  }
  
  /**
   * Sends a PING message to the IRC server.
   *
   * @param   server1   The server that must be pinged.
   */
  public void sendPINGMessage(String server1) {
    sendRawMessage("PING " + server1);
  }
  
  /**
   * Sends a PING message to the IRC server.
   *
   * @param   server1   The server that must be pinged.
   * @param   server2   The PING will be forwarded to this server.
   */
  public void sendPINGMessage(String server1, String server2) {
    sendRawMessage("PING " + server1 + " " + server2);
  }
  
  /**
   * Sends a PONG message to the IRC server.
   *
   * @param   daemon    Name of the daemon that has responded.
   */
  public void sendPONGMessage(String daemon) {
    sendRawMessage("PONG " + daemon);
  }
  
  /**
   * Sends a PONG message to the IRC server.
   *
   * @param   daemon    Name of the daemon that has responded.
   * @param   daemon2   The PONG message will be forwarded to daemon2.
   */
  public void sendPONGMessage(String daemon, String daemon2) {
    sendRawMessage("PONG " + daemon + " " + daemon2);
  }
  
  /**
   * Sends a JOIN message to the IRC server. With it the user can try to join
   * a channel (or more). If more than one channels should be joined they should
   * be comma separated, and there must not be spaces between their names.
   *
   * @param   channels    The channels that must be joined
   */
  public void sendJOINMessage(String channels) {
    sendRawMessage("JOIN " + channels);
  }
  
  /**
   * Sends a JOIN message to the IRC server. With it the user can try to join
   * a channel (or more). If more than one channels should be joined they should
   * be comma separated, and there must not be spaces between their names. If there are
   * more than one key they also must be comma separated and without spaces.
   *
   * @param   channels    The channels that must be joined
   * @param   keys        The keys to enter the channels (they are optional).
   */
  public void sendJOINMessage(String channels, String keys) {
    sendRawMessage("JOIN " + channels + " " + keys);
  }
  
  /**
   * Sends a PART message to the IRC server. If the user wants to leave more than one
   * channel the names of them should be comma separated and without spaces.
   *
   * @param   channels    The channels that must be parted.
   */
  public void sendPARTMessage(String channels) {
    sendRawMessage("PART " + channels);
  }
  
  /**
   * Sends a TOPIC message to the IRC server to retrieve the topic of a channel.
   *
   * @param   channelName   The name of the channel that the topic should be retrieved.
   */
  public void sendTOPICMessage(String channelName) {
    sendRawMessage("TOPIC " + channelName);
  }
  
  /**
   * Sends a TOPIC message to the IRC server to set the topic of a channel.
   *
   * @param   channelName   The name of the channel that the topic will be set.
   * @param   topic         The desired topic for the channel.
   */
  public void sendTOPICMessage(String channelName, String topic) {
    sendRawMessage("TOPIC " + channelName + " :" + topic);
  }
  
  /**
   * Sends a PRIVMSG message to the IRC server. The receiver can be channels and / or nicks.
   * To specify more than one receiver they must be comma separated with no spaces between
   * them.
   *
   * @param   receivers     The receivers of the PRIVMSG.
   * @param   message       The text of the message.
   */
  public void sendPRIVMSGMessage(String receivers, String message) {
    sendRawMessage("PRIVMSG " + receivers + " :" + message); 
  }
  
  /**
   * Sends an <code>IRCMessage</code> to the IRC server.
   *
   * @param   msg   The message to be sended.
   */
  public void sendMessage(IRCMessage msg) {
    sendRawMessage(msg.getPrefix() + " " + msg.getCommand() + " " + msg.getParams());
  }
  
  /**
   * Sends a raw message to the IRC server. It should not include the final CRLF,
   * although there will be no problem if the message includes them (a void message
   * will be sended).
   */
  public void sendRawMessage(String msg) {
    if (isClosed()) {
      System.err.println("Cannot send message: no connection");
    } else {
      if (msg.length() > 510) {
        msg = msg.substring(0, 510);
      }
      
      boolean correct = connection.write((msg + "\r\n").getBytes());
      
      if (correct) {
        //        System.err.println("Send --> " + msg);
      } else {
        disconnect();
      }
    }
  }
}
