/**
 * This file is yet under development. Please use at your own risk.
 *
 * Zerjillo
 */
package com.zerjio.JIRC;

/**
 * This class represents a basic message that is send over an IRC Connection.
 * <P>
 * It is recommended to take a look to the RFC 1459 to know what kinds of messages
 * do exist and their exact formats. All IRC Messages should be a subclass of this
 * generic one.
 *
 * @author  Zerjillo (http://www.zerjio.com)
 * @version 0.9b
 */
public class IRCMessage {
  /** Prefix of the IRC Message */
  private String prefix;
  /** Nick (contained in the prefix) */
  private String nick;
  /** User (contained in the prefix) */
  private String user;
  /** Host (contained in the prefix) */
  private String host;
  
  /** Command of the IRC Message */
  private String command;
  /** Parameters of the IRC Message */
  private String params;
  /** Number of times that the message has been "used"
   *  @see isUsed()
   *  @see markAsUsed()
   *  @see getUsedCount()
   */
  private int usedCount;
  /** Marks if the message must be discarded (no other <code>MsgListener</code>
   *  will have access to it).
   *  @see mustBeDiscarded()
   *  @see getMustBeDiscarded()
   */
  private boolean mustBeDiscarded;
  
  /**
   * Constructs a general IRC message with a prefix, a command and its parameters.
   *
   * @param   prefix    The IRC message prefix.
   * @param   command   The IRC message command.
   * @param   params    The IRC message parameters.
   */
  public IRCMessage(String prefix, String command, String params) {
    if (prefix == null) {
      this.prefix = "";
    } else {
      this.prefix = prefix;
      
      getInfoFromPrefix();
    }
    
    if (command == null) {
      this.command = "";
    } else {
      this.command = command;
    }
    
    if (params == null) {
      this.params = "";
    } else {
      this.params = params;
    }
    
    usedCount = 0;
    mustBeDiscarded = false;
  }
  
  /**
   * Gets The nick / server, the user and host from the prefix.
   */
  private void getInfoFromPrefix() {
    int exclamationPos = prefix.indexOf('!');
    int arrobaPos = prefix.indexOf('@');
    
    if ((exclamationPos != -1) && (arrobaPos != -1)) {
      nick = prefix.substring(1, exclamationPos);
      user = prefix.substring(exclamationPos + 1, arrobaPos);
      host = prefix.substring(arrobaPos + 1);
    } else if ((exclamationPos != -1) && (arrobaPos  == -1)) {
      nick = prefix.substring(1, exclamationPos);
      user = prefix.substring(exclamationPos + 1);
      host = "";
    } else if ((exclamationPos == -1) && (arrobaPos != -1)) {
      nick = prefix.substring(1, arrobaPos);
      user = "";
      host = prefix.substring(arrobaPos + 1);
    } else {
      if (prefix.length() > 1) {
        nick = prefix.substring(1);
      } else {
        nick = "";
      }
      user = "";
      host = "";
    }
  }
  
  /**
   * Gets the IRC message prefix.
   *
   * @return  The IRC message prefix.
   */
  public String getPrefix() {
    return prefix;
  }
  
  /**
   * Gets the nick (part of prefix).
   *
   * @return  The nick.
   */
  public String getNick() {
    return nick;
  }
  
  /**
   * Gets the user (part of prefix).
   *
   * @return  The user.
   */
  public String getUser() {
    return user;
  }
  
  /**
   * Gets the host (part of prefix).
   *
   * @return  The host.
   */
  public String getHost() {
    return host;
  }
  
  /**
   * Gets the IRC message command.
   *
   * @return  The IRC message command.
   */
  public String getCommand() {
    return command;
  }
  
  /**
   * Gets the IRC message parameters.
   *
   * @return  The IRC message parameters.
   */
  public String getParams() {
    return params;
  }
  
  /**
   * Gets how many times have this IRC message been used.
   *
   * @return  The number of times that the IRC message has been used.
   */
  public int getUsedCount() {
    return usedCount;
  }
  
  /**
   * Returns if the IRC message has been previously used.
   *
   * @return  <code>true</code> if the IRC message has been previously used.
   *          <code>false</code> if the IRC message has NOT been previously used.
   */
  public boolean isUsed() {
    return (usedCount != 0);
  }
  
  /**
   * Marks the IRC message as used. To do so it increments the usedCount counter.
   */
  public void markAsUsed() {
    usedCount++;
  }
  
  /**
   * Marks the IRC message to be discarded. That should avoid that another message
   * listeners to access it.
   */
  public void mustBeDiscarded() {
    mustBeDiscarded = true;
  }
  
  /**
   * Gets if the IRC message must be discarded.
   *
   * @return    <code>true</code> if the IRC message must be discarded.
   *            <code>false</code> if NOT must be discarded.
   */
  public boolean getMustBeDiscarded() {
    return mustBeDiscarded;
  }
  
  /**
   * Gives a String representation of the IRC message. Usually the format correspond
   * to the exact format in what has been sent.
   *
   * @return  A String that represents this IRC message.
   */
  public String asString() {
    return (prefix + " " + command + " " + params).trim();
  }
  
  /**
   * Gives a String representation of the IRC message but ignoring its prefix.
   *
   * @return  A String that represents this IRC message without prefix.
   */
  public String asStringWithoutPrefix() {
    return (command + " " + params).trim();
  }
}
